;;; basic doc parser -- do not emulate this terrible, terrible code

(use regex posix srfi-13)

(define (process-line line)
  (cond ((blank-line? line) '())
        ((eqv? (string-ref line 0) #\;)
         (cond ((string-search "^;;; " line)
                (cons 'section
                      (cons (transform-text (strip-prefix ";;; " line))
                            (read-section ";;;;+ "))))
               ((string-search "^;;;; " line)
                (cons 'subsection
                      (cons (transform-text (strip-prefix ";;;; " line))
                            (read-section ";;;;;+ "))))
               ((string-search "^;; " line)
                (cons 'p (transform-text (read-paragraph ";; " line))))
               ((string-search "^;< " line)
                (list 'pre (read-code-paragraph ";[<;] " line)))
               ((string-search "^;R " line)
                (list 'repl (read-code-paragraph ";[R;] " line)))
               (else
                (list 'pre (read-code-paragraph "" line)))))
        (else
         (list 'pre (read-code-paragraph "" line)))))

(define (strip-prefix prefix str)
  (if (string=? prefix "")         ; string-substitute on null or ^ regexp hangs
      str
      (string-substitute (string-append "^" prefix)
                         ""
                         str)))

(define (blank-line? line)
  (string-search "^(#\\||\\|#)?\\s*$" line))   ;; ignore blank lines and #| and |#

(define (read-paragraph prefix first-line . joiner)
  (let next-line ((para (strip-prefix prefix first-line)))
    (let ((line (read-line)))
      (if (or (eof-object? line)
              (blank-line? line))
          para
          (next-line (string-append para (:optional joiner " ")
                                    (strip-prefix prefix line)))))))

(define (read-code-paragraph prefix first-line)
  (read-paragraph prefix first-line "\n"))

(define (read-section subsection-prefix)
  (process-doc (lambda (line)
                 (or (eof-object? line)
                     (and (string-search "^;;;+" line) ;; this is a section
                          (not (string-search ;; but not a subsection of ours
                                (string-append "^" subsection-prefix) line)))))))

;; (read-paragraph (open-input-string ";; line two\n;; line three\n  ") ";; " ";; line one")
;; => line one line two line three

(define (parse-doc filename)
  (with-input-from-file filename
    init-links)
  (with-input-from-file filename
    (lambda () (process-doc (lambda (x) (or (string=? *links-marker* x) ;; end at section Links
                                            (eof-object? x)))))))

(define (%doit)   ; internal for testing
  (write-list
   (parse-doc "~/scheme/xcode/plist/itunes.scm")
   "/tmp/itunes-docparse.scm"))

(define (process-doc . end?)
  (let ((end? (:optional end? eof-object?)))
   (reverse
    (let loop ((result '()))
      (let ((line (peek-line)))
        (if (end? line)
            result
            (begin (read-line)  ; discard
             (let ((ret (process-line line)))
               (if (pair? ret)
                   (loop (cons ret result))
                   (loop result))))))))))

;; doesn't work on string input
(define (peek-line . port)
  (let ((port (:optional port (current-input-port))))
    
    (let ((pos (file-position port))
          (line (read-line port)))
      (set-file-position! port pos)
      line)))


;; Will result in empty strings if delimiters begin or end the
;; entire string, but this doesn't matter for SXML.
;; Does not handle nested tags correctly.  Does not handle
;; backslashed characters correctly.
(define not-backslash "(?<!\\\\)")        ;; zero-width negative lookbehind for \
(define backslash "\\\\")

;; Unescape [using global value of backslash] CHARS in STR.
(define (unescape str . chars)
  (if (null? chars)
      str
      (string-substitute (string-append backslash
                                        "("
                                        (string-join (map string chars) "|")
                                        ")")
                         "\\1"
                         str 'global)))

(define (split-delimiters open close str tagger)
  (let ((regexp (string-append not-backslash
                               (string open)
                               "(?:.*?)"
                               not-backslash
                               (string close))))
    (split-regexp regexp str tagger
                  (lambda args (unescape (apply substring args)
                                         open close)))))

(define (split-regexp regexp str tagger . substr)
  (let ((substring (:optional substr substring)))
   (let loop ((pos 0))
     (let ((positions (string-search-positions regexp str pos)))
       (if (not positions)
           (list (substring str pos))
           (let ((start (caar  positions))
                 (end   (cadar positions)))
             (cons (substring str pos start)
                   (cons (tagger (substring str (fx+ start 1) (fx- end 1)))
                         (loop end)))))))))

(define (split-mdash str)
  (split-regexp (string-append not-backslash
                               "--+")
                str
                (lambda (x) (list '& "mdash"))
                (lambda args (unescape (apply substring args)
                                       #\-))))

(define (split-links str)
  (split-delimiters #\{ #\} str
                    (lambda (link)
                      `(url ,(lookup-link link) ,link))))

;; (split-delimiters #\< #\> "the <quick> brown <fox> jumped" (lambda (x) (list 'tt x)))
;; => ("the " (tt "quick") " brown " (tt "fox") " jumped")

;; It occurs to me we could do this transformation using SXSLT.
(define (transform-text str)
  (transform-list
   (transform-list
    (split-delimiters #\< #\> str (lambda (x) (list 'tt x)))
    (lambda (str) (split-mdash str)))
   (lambda (str) (split-links str))))

(define (transform-list L transformer)
  (if (null? L) '()
      (cond ((pair? (car L))
             (cons (transform-list (car L) transformer)
                   (transform-list (cdr L) transformer)))
            ((string? (car L))
             (append (transformer (car L))    ;; cons is faster and accepted by eggdoc, though
                     (transform-list (cdr L) transformer)))
            (else (cons (car L)
                        (transform-list (cdr L) transformer))))))


;;; Links

(define *global-link-table*)
(define *links-marker* ";;; Links")
(define (lookup-link link)
  (hash-table-ref *global-link-table* link
                  (lambda ()
                    (warning (conc "invalid link referent: " link))
                    "")))

(define (init-links)
  ;; Just skip until Links section is reached.
  (set! *global-link-table* (make-hash-table string=?))
  (let loop ()
    (let ((line (read-line)))
      (if (or (eof-object? line)
              (string=? line *links-marker*))
          'done
          (loop))))

  (let loop ()
    (let ((line (read-line)))
      (if (eof-object? line)
          'done
          (let ((matches (string-search "^;; {(.*?)}\\s*([^\\s]+)" line)))
            (if matches
                (let ((key (cadr matches))
                      (url (caddr matches)))
                  (hash-table-set! *global-link-table* key url)
                  (loop))
                (loop)))))))